/**********************************************************************
Copyright (c) 2010 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.jpa;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.util.Elements;
import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.MappedSuperclass;
import javax.tools.FileObject;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.StandardLocation;
import javax.tools.ToolProvider;

/**
 * Annotation processor for JPA to generate "dummy" classes for all Entity classes for use with Criteria API.
 * Any class ({MyClass}) that has a JPA "class" annotation will have a stub class ({MyClass}_) generated.
 */
@SupportedAnnotationTypes({"javax.persistence.Entity", "javax.persistence.Embeddable", "javax.persistence.MappedSuperclass"})
@SupportedSourceVersion(SourceVersion.RELEASE_6)
public class JPACriteriaProcessor extends AbstractProcessor
{
    final Set<File> locations = new HashSet<File>();
    final Set<String> classNames = new HashSet<String>();

    /* (non-Javadoc)
     * @see javax.annotation.processing.AbstractProcessor#process(java.util.Set, javax.annotation.processing.RoundEnvironment)
     */
    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv)
    {
        if (roundEnv.processingOver())
        {
            return false;
        }

        Set<? extends Element> elements = roundEnv.getRootElements();
        for (Element e : elements)
        {
            process((TypeElement)e);
        }
        return false;
    }

    /**
     * Handler for processing a JPA annotated class to create the criteria class stub.
     * @param element The class element
     */
    protected void process(TypeElement element)
    {
        if (element == null || !isJPAAnnotated(element))
        {
            return;
        }

        Elements elementUtils = processingEnv.getElementUtils();
        String originalClass = elementUtils.getBinaryName(element).toString();
        String newClassName = originalClass + "_";
        // TODO Find location to write classes to so they are alongside the input source classes
        System.out.println("JPA2 Criteria Processor original=" + originalClass + " new=" + newClassName);

        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager fileManager = compiler.getStandardFileManager(null, null, null);
        try
        {
            JavaFileObject javaFile = fileManager.getJavaFileForOutput(StandardLocation.SOURCE_OUTPUT,
                newClassName, JavaFileObject.Kind.CLASS, null);

            OutputStream out = javaFile.openOutputStream();
            // TODO Add contents of source file
            out.close();
        }
        catch (IOException ioe)
        {
        }
    }

    protected String getClassRoot()
    {
        // get classpath roots (classoutput, sourceoutput, source path, or current dir
        String classpath = "";
        try
        {
            FileObject f = processingEnv.getFiler().getResource(StandardLocation.CLASS_OUTPUT, "", "d");
            classpath = new File(new URL(f.toUri().toString()).toURI()).getParentFile().getAbsolutePath() + File.separatorChar;
        }
        catch (Exception e1)
        {
            try
            {
                FileObject f = processingEnv.getFiler().getResource(StandardLocation.SOURCE_OUTPUT, "", "d");
                classpath = new File(new URL(f.toUri().toString()).toURI()).getParentFile().getAbsolutePath() + File.separatorChar;
            }
            catch (Exception e2)
            {
                try
                {
                    FileObject f = processingEnv.getFiler().getResource(StandardLocation.SOURCE_PATH, "", "d");
                    classpath = new File(new URL(f.toUri().toString()).toURI()).getParentFile().getAbsolutePath() + File.separatorChar;
                }
                catch (Exception e3)
                {
                    try
                    {
                        classpath = System.getProperty("user.dir") + File.separatorChar;
                    }
                    catch (Exception e4)
                    {
                    }
                }
            }
        }
        return classpath;
    }

    /**
     * Convenience method to return if this class element has any of the defining JPA annotations.
     * @param el The class element
     * @return Whether it is to be considered a JPA annotated class
     */
    protected boolean isJPAAnnotated(Element el)
    {
        if ((el.getAnnotation(Entity.class) != null) ||
            (el.getAnnotation(MappedSuperclass.class) != null) ||
            (el.getAnnotation(Embeddable.class) != null))
        {
            return true;
        }
        return false;
    }
}